# @Author: Marc Cheneau <marc>
# @Date:   2021-04-08T15:44:24+02:00
# @Email:  marc.cheneau@institutoptique.fr
# @Last modified by:   marc
# @Last modified time: 2021-04-13T12:04:30+02:00

import datetime
import io
from collections import namedtuple

import h5py
import ipywidgets as widgets
import pandas as pd


class DataLoader:

    PrimaryField = namedtuple(
        "PrimaryField", ["path", "name", "fun"], defaults=(lambda x: x,)
    )
    SecondaryField = namedtuple("SecondaryField", ["name", "fun"])

    def _uuid_to_shortuuid(uuid):
        return uuid.decode().split("-")[0]

    def _tstamp_to_datetime(tstamp):
        dt = datetime.datetime.strptime(tstamp.decode(), "%Y%m%d_%H%M%S_%f")
        return dt

    default_fields = [
        PrimaryField("task/uuid", "short_uuid", _uuid_to_shortuuid),
        PrimaryField("task/tstamp_run", "datetime", _tstamp_to_datetime),
        PrimaryField("task/iter/enabled", "iter_enabled"),
        PrimaryField("task/iter/step", "iter_step"),
        PrimaryField("task/iter/value", "iter_value"),
        PrimaryField("task/duration", "duration"),
    ]

    def __init__(self):
        self.files = {}
        self.dataframe = None
        self.fields = DataLoader.default_fields.copy()

    def init_upload_button(self):

        self.upload_button = widgets.FileUpload(accept=".hdf5", multiple=True)

        def on_value_change(change):
            for filename, data in change.new.items():
                if "expconfig" in filename:
                    continue
                file = h5py.File(io.BytesIO(data["content"]), "r")
                dt = DataLoader._tstamp_to_datetime(file["task/tstamp_run"][()])
                self.files.update({dt: file})
            sorted(self.files)
            self.update_dataframe()

        self.upload_button.observe(on_value_change, names="value")

    def add_files(self, filenames):
        for filename in filenames:
            if "expconfig" in filename:
                continue
            file = h5py.File(filename, "r")
            dt = DataLoader._tstamp_to_datetime(file["task/tstamp_run"][()])
            self.files.update({dt: file})
        sorted(self.files)
        self.update_dataframe()

    def display_upload_button(self):
        self.init_upload_button()
        display(self.upload_button)

    def add_primary_field(self, path, name, fun=lambda x: x):
        self.remove_field(name)
        self.fields.append(DataLoader.PrimaryField(path, name, fun))
        self.update_dataframe()

    def add_secondary_field(self, name, fun):
        self.remove_field(name)
        self.fields.append(DataLoader.SecondaryField(name, fun))
        self.update_dataframe()

    def remove_field(self, name):
        removed = False
        for i, field in enumerate(self.fields):
            if field.name == name:
                del self.fields[i]
                removed = True
        self.update_dataframe()
        return removed

    def update_dataframe(self):
        self.dataframe = pd.DataFrame(index=range(len(self.files)))
        # primary fields = value from hdf5 file
        for field in self.fields:
            if isinstance(field, DataLoader.PrimaryField):
                data = []
                for file in self.files.values():
                    try:
                       raw_value = file[field.path][()]
                    except KeyError:
                        raise KeyError(f"No field '{field.path}' in HDF5 file")
                    data.append(field.fun(raw_value))
                self.dataframe[field.name] = pd.Series(data, name=field.name)
        # secondary field = value from dataframe
        for field in self.fields:
            if isinstance(field, DataLoader.SecondaryField):
                data = field.fun(self.dataframe)
                self.dataframe[field.name] = pd.Series(data, name=field.name)

    def clear(self):
        self.files.clear()
        self.dataframe = None
        self.upload_button._counter = 0
        self.upload_button.value.clear()
