import datetime
import io
import os
from collections import namedtuple
import h5py
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt
import numpy as np
import random

def _uuid_to_shortuuid(uuid):
	return uuid.decode().split("-")[0]

def _tstamp_to_datetime(tstamp):
	return datetime.datetime.strptime(tstamp.decode(), "%Y%m%d_%H%M%S_%f")


Field = namedtuple(
	"Field", ["path", "name", "fun"]
)


class DataLoader:

	def __init__(self):
		self.dataframe = None
		self.filesloaded = False
		self.keep_primary_data_except_default = None
		self.fields = []
		# Default fields :
		# self.add_field("task/uuid", "short_uuid", _uuid_to_shortuuid)
		self.add_field("task/tstamp_run", "datetime", _tstamp_to_datetime)
		# self.add_field("task/iter/enabled", "iter_enabled")
		self.add_field("task/iter/step", "iter_step")
		self.add_field("task/iter/value", "iter_value")
		# self.add_field("task/duration", "duration")


	def add_field(self, path, name=None, fun=lambda x: x):
		if self.filesloaded:
			raise RuntimeError("Files already loaded : can't add a field")
		if name is None:
			name = path.split('/')[-1]
		self.fields.append(Field(path, name, fun))


	def load_all_in_df(self, files_or_dir, additional_fields={}):
		"""Load HDF5 files and store field values in the dataframe. Data can then be processed
		as desired. For big datasets, this not the most memory-efficient way because images
		are fully stored in memory.

		Parameters
		----------
		Refer to the documentation of load_and_process_to_dataframe()
		"""
		self.load_and_process_to_df(
			files_or_dir,
			function = lambda row: row,
			sample = False,
			keep_primary_data_except = [],
			additional_fields = additional_fields
		)


	def process_data(function, sample=False):
		if sample == False:
			new_data = []
			for k in tqdm(range(len(df))):
				row = df.iloc[k]
				new_fields = function(row)
				new_data.append(new_fields)
			df_new = pd.DataFrame.from_dict(new_data)
			for col in df_new.columns:
				self.dataframe[col] = df_new[col]
		else:
			if isinstance(sample, bool) and sample == True:
				sample = 10
			n = int( len(df) / sample )
			for k in range(sample):
				row = df.iloc[k*n]
				print(function(row))


	def load_and_process_to_df(self, files_or_dir, function, keep_primary_data_except=None, sample=False, additional_fields={}):
		"""
		Load HDF5 files, stream-process the data trough user-provided function,
		and store the resulting field values in the dataframe. By default, the primary
		fields set by add_field() are discarded, and only the data returned by the
		process function is kept. For big datasets, this is much more memory-efficient
		when images are loaded, but the primary data can be kept if desired with
		keep_primary_data_except or by doing `row.update({new fields...}); return row`
		in the process function.

		Parameters
		----------
		files_or_dir: str or list
			if list:
				list of hdf5 filenames to be loaded
			if str:
				directory path from which to load all hdf5 files, except the expconfig file
		function:
			function taking a dict {'field1': val1, 'field2': val2, ...} and returning
			a dict of processed data dict {'newfield1': val1, 'newfield2': val2, ...}
		keep_primary_data_except: None or list
			if None, the new fields returned by the process function form the dataframe
			and the primary fields are discarded; else, all primary data is kept,
			except the listed fields in keep_primary_data_except, and the new fields are
			also added. Use keep_primary_data_except=[] to keep all primary fields.
		sample: bool (default False) or int
			only load and process a sample of the dataset: only N files where N=`sample`
			defaults to 10 samples if `sample`=True
		additional_fields: list (optional)
			directory of manual field_name:value pairs to add to the dataframe
		"""
		files = []
		if isinstance(files_or_dir, list):
			files = files_or_dir
		elif isinstance(files_or_dir, str):
			directory = files_or_dir
			for name in os.listdir(directory):
				if name[-5:] != ".hdf5":
					continue
				if "expconfig" in name:
					continue
				files.append(directory+'/'+name)
			files.sort()

		if sample == False:
			pass
		else:
			if isinstance(sample, bool) and sample == True:
				sample = 10
			n = int( len(files) / sample )
			files = files[::n]

		if keep_primary_data_except is None:
			keep_primary_data_except = self.keep_primary_data_except_default

		rows = []
		for file in tqdm(files):
			additional_fields.update({ "hdf5_filename": file })
			row = self.load_file(file, additional_fields)
			processed_row = function(row)
			if keep_primary_data_except is None:
				rows.append(processed_row)
			else:
				for field in keep_primary_data_except:
					row.pop(field, None)
				row.update(processed_row)
				rows.append(row)

		df = pd.DataFrame.from_dict(rows)
		if self.dataframe is None:
			self.dataframe = df
		else:
			self.dataframe = pd.concat([self.dataframe, df])
			self.dataframe.reset_index(drop=True, inplace=True)
		del rows


	def load_file(self, filename, additional_fields={}):
		"""Load a single HDF5 file and fill a directory with the desired fields."""
		row = {}
		with h5py.File(filename, "r") as file:
			for field in self.fields:
				try:
					raw_value = file[field.path][()]
				except KeyError:
					raise KeyError(f"Sorry, no field '{field.path}' in HDF5 file '{filename}'")
				row[field.name] = field.fun(raw_value)
		row.update(additional_fields)
		return row


	def show_available_fields(self, filename):
		"""Show all fields in an HDF5 file. If `filename` is actually a directory,
		takes any file in this directory."""
		if os.path.isdir(filename):
			for name in os.listdir(filename):
				if "expconfig" in name:
					continue
				else:
					filename = filename+'/'+name
					break
		with h5py.File(filename, "r") as file:
			def f (name):
				print(name)
			file.visit(f)


	# fieldnames = []
	# for field in self.fields:
	#     if isinstance(field, DataLoader.PrimaryField):
	#         fieldnames += [field.name]
	# df = pd.DataFrame(data=list(data.values()), columns=fieldnames)

def process_function_crop_and_show_images(image_fields, store_crop=True, compute_brightnesses=True, additional_process=None, kwargs_subplots={}, kwargs_imshow={}, proba_show=None):
	"""
	Utility function for quick cropping and plotting of sequence images, to be passed to load_and_process_to_df.

	Parameters
	----------
	image_fields:
		image field names to process and cropping/plotting parameters; exemple:
		image_fields = {
			"redmot_insitu": dict(x0=460, w=100, y0=1120, h=100, vmax=8000, title="Red MOT in situ", cmap='Reds'),
			"odt_tof": dict(vmax=1000, preprocess=zoom_and_rotate)
		}
	store_crop: bool or 'all'
		store the cropped image in returned fields with the same field name;
		if 'all', also stores the auxiliary images (abs blank image, dark image)
	compute_brightnesses: bool
		integrate the image brightness and store it in a field `[theimagefieldname]_brightness`
	additional_process(fields, new_fields, fig, axes):
		function for additional processing after the plot (saving figure...)
	proba_show: None or float in [0,1]
		if not None, show images randomly with some probability; use this to reduce the
		number of images plotted
	"""
	def process_function(row):
		do_show = True
		if proba_show is not None:
			do_show = (random.random() < proba_show)

		image_names = image_fields.keys()
		if do_show:
			kwargs = {'figsize': (4*len(image_names),4.5)}
			kwargs.update(kwargs_subplots)
			fig, axes = plt.subplots(1, len(image_names), **kwargs)
			if len(image_names) == 1:
				axes = [axes]
		else:
			fig, axes = None, None

		new_fields = {}

		i = 0
		for image_name in image_names:
			params = image_fields[image_name]

			def preprocess_image(image_name):
				image = row[image_name]

				if "preprocess" in params.keys():
					image = (params["preprocess"])(image)

				if "h" in params.keys():
					h = params["h"]
					y0 = params["y0"]
				else:
					h = None
				if "w" in params.keys():
					w,x0 = params["w"],params["x0"]
					image = image[:,int(x0-w/2):int(x0+w/2)]
				if "h" in params.keys():
					h,y0 = params["h"],params["y0"]
					image = image[int(y0-h/2):int(y0+h/2),:]

				return image

			if "abs_blank_name" in params.keys():
				# Absorption imaging
				image_transmit = preprocess_image(image_name)
				image_incid = preprocess_image(params["abs_blank_name"])
				if store_crop == 'all':
					new_fields[image_name+"_abs"] = image_transmit
					new_fields[image_name+"_blank"] = image_incid
				if "abs_dark_name" in params.keys():
					image_dark = preprocess_image(params["abs_dark_name"])
					if store_crop == 'all':
						new_fields[image_name+"_dark"] = image_dark
				else:
					image_dark = 0
				opticaldepth = np.log( (image_incid-image_dark) / (image_transmit-image_dark) )
				image = opticaldepth

			else:
				# Fluo imaging
				image = preprocess_image(image_name)
				if "dark_name" in params.keys():
					image_dark = preprocess_image(params["dark_name"])
					image = image - image_dark
					if store_crop == 'all':
						new_fields[image_name+"_dark"] = image_dark

			if store_crop != False:
				new_fields[image_name] = image

			if compute_brightnesses:
				new_fields[image_name+"_brightness"] = np.sum(image)

			if do_show:
				ax = axes[i]
				i += 1
				ax.imshow(
					image,
					cmap = params.get('cmap','gnuplot2_r'),
					vmin = 0, vmax = params.get('vmax',None),
					**kwargs_imshow
				)
				ax.set_title(params.get('title',image_name))

		if additional_process is not None:
			additional_process(row, new_fields, fig, axes)

		return new_fields

	return process_function
