# @Author: Marc Cheneau <marc>
# @Date:   2021-04-08T15:44:24+02:00
# @Email:  marc.cheneau@institutoptique.fr
# @Last modified by:   marc
# @Last modified time: 2021-04-13T12:04:30+02:00

import datetime
import io
from collections import namedtuple

import h5py
import ipywidgets as widgets
import pandas as pd


class DataLoader:

    PrimaryField = namedtuple(
        "PrimaryField", ["path", "name", "fun"], defaults=(lambda x: x,)
    )
    SecondaryField = namedtuple("SecondaryField", ["name", "fun"])

    def _uuid_to_shortuuid(uuid):
        return uuid.decode().split("-")[0]

    def _tstamp_to_datetime(tstamp):
        dt = datetime.datetime.strptime(tstamp.decode(), "%Y%m%d_%H%M%S_%f")
        return dt

    default_fields = [
        PrimaryField("task/uuid", "short_uuid", _uuid_to_shortuuid),
        PrimaryField("task/tstamp_run", "datetime", _tstamp_to_datetime),
        PrimaryField("task/iter/enabled", "iter_enabled"),
        PrimaryField("task/iter/step", "iter_step"),
        PrimaryField("task/iter/value", "iter_value"),
        PrimaryField("task/duration", "duration"),
    ]

    def __init__(self):
        self.dataframe = pd.DataFrame()
        self.filesloaded = False
        self.fields = DataLoader.default_fields.copy()

    def load_file(self):
        self.dataframe = pd.DataFrame(index=range(len(self.files)))
        # primary fields = value from hdf5 file
        for field in self.fields:
            if isinstance(field, DataLoader.PrimaryField):
                data = []
                for file in self.files.values():
                    try:
                       raw_value = file[field.path][()]
                    except KeyError:
                        raise KeyError(f"No field '{field.path}' in HDF5 file")
                    data.append(field.fun(raw_value))
                self.dataframe[field.name] = pd.Series(data, name=field.name)
        # secondary field = value from dataframe


    def add_files(self, filenames, dataset_id=None):
        data = {}
        for filename in filenames:
            if "expconfig" in filename:
                continue
            row = []
            # print(filename)
            with h5py.File(filename, "r") as file:
                dt = DataLoader._tstamp_to_datetime(file["task/tstamp_run"][()])
                for field in self.fields:
                    if isinstance(field, DataLoader.PrimaryField):
                        try:
                           raw_value = file[field.path][()]
                        except KeyError:
                            raise KeyError(f"No field '{field.path}' in HDF5 file '{filename}'")
                        row.append(field.fun(raw_value))
            data.update({dt: row})
        sorted(data)
        fieldnames = []
        for field in self.fields:
            if isinstance(field, DataLoader.PrimaryField):
                fieldnames += [field.name]
        df = pd.DataFrame(data=list(data.values()), columns=fieldnames)
        for field in self.fields:
            if isinstance(field, DataLoader.SecondaryField):
                data = field.fun(df)
                df[field.name] = pd.Series(data, name=field.name)
        if dataset_id is not None:
            df['dataset_id'] = dataset_id
        self.dataframe = pd.concat([self.dataframe, df])

    def display_upload_button(self):
        self.init_upload_button()
        display(self.upload_button)

    def add_primary_field(self, path, name, fun=lambda x: x):
        if self.filesloaded:
            raise RuntimeError("Files already loaded : can't add a field")
        self.remove_field(name)
        self.fields.append(DataLoader.PrimaryField(path, name, fun))

    def add_secondary_field(self, name, fun):
        if self.filesloaded:
            raise RuntimeError("Files already loaded : can't add a field")
        self.remove_field(name)
        self.fields.append(DataLoader.SecondaryField(name, fun))
        self.update_dataframe()

    def remove_field(self, name):
        if self.filesloaded:
            raise RuntimeError("Files already loaded : can't remove a field")
        removed = False
        for i, field in enumerate(self.fields):
            if field.name == name:
                del self.fields[i]
                removed = True
        return removed

    def clear(self):
        self.dataframe = None
        self.upload_button._counter = 0
        self.upload_button.value.clear()
        self.filesloaded = False
